%%%%%%%%%%%%%%%%%%%%%%%%%%%%% -*- Mode: Prolog -*- %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% precompile.pl --- 
%% Author          : Massimo Fasciano
%% Last Modified On: Fri Oct 22 17:41:44 1993
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- ensure_loaded('include.pl').
:- ensure_loaded('tools.pl').
:- ensure_loaded('unif.pl').

% precompiles a PFUF grammar file into a Prolog version
% that the unifier can handle...
precompile(FileIN,FileOUT) :-
    see(FileIN),lire_fichier(L),seen,
    trans(L,L2),optimize_program(L2,L3),
    tell(FileOUT),pr_prog(L3),told.

% variables are left unchanged.
tr(Var,Top,Var,Top) :- var(Var),!.
% translate from feature-level disjunction to rule-level disjunction.
tr((A===BorC),Top,P,Top2) :-
    nonvar(BorC),BorC=(B;C),!,
    featureDisj((A===BorC),RuleLevel),tr(RuleLevel,Top,P,Top2).
% translate FD unification.
tr((A===B),Top,TR,Top) :-
    !,tra((A===B),TR).
% translate a rule
% perform simple unifications at compile-time to make indexing possible.
tr((A:-(X=Y,B)),Top,T,Top2) :- var(X),nonvar(Y),X=Y,!,tr((A:-B),Top,T,Top2).
tr((A:-(X=Y)),Top,A,Top) :- var(X),nonvar(Y),X=Y,!.
tr((A:-B),Top,Rule,Top2) :-
    !,tr(B,Top,BT,Top2),(BT=true,!,Rule=A;Rule=(A:-BT)).
% translate the given/1 predicate
tr(given(A),Top,(Unif,nonvar(Val)),Top) :-
    !,tra((A === Val),Unif).
% translate ";"-based disjunction into a new predicate
% with alternate rules.
tr((A;B),Top,Proc,Top2) :-
    !,makeRules((A;B),Proc,Rules),conc(Rules,Top,Top2).
% translate conjunction.
tr((A,B),Top,T,Top3) :-
    !,tr(A,Top,AT,Top2),tr(B,Top2,BT,Top3),conjTrue(AT,BT,T).
% catch-all case.
tr(A,Top,A,Top).

% translate a list of rules
trans([],[]).
trans([H1|T1],[H2|T2]) :- tr(H1,T1,H2,T3),trans(T3,T2).

% translate "==="-based unification into "unif"-based unification.
tra((A1 === A2),(unif(OS1,L1),unif(OS2,L2))) :-
  nonvar(A1),nonvar(A2),
  A1 = L1:P1,A2 = L2:P2,
  !,cvtOp(P1,S1,Val),cvtOp(P2,S2,Val),
    ouverte(S1,OS1),ouverte(S2,OS2).
tra((A1 === A2), unif(OS1,L1)) :-
  (nonvar(A1),A1 = L1:P1,A2 = Val,!;
   nonvar(A2),A2 = L1:P1,A1 = Val),
  !,cvtOp(P1,S1,Val),ouverte(S1,OS1).
tra((L1 === L2),unif(OL1,OL2)) :-
    ouverte(L1,OL1),ouverte(L2,OL2),!.

cvtOp((A:B),[A:L|_],Val) :- !,cvtOp(B,L,Val).
cvtOp(X,[X:Val|_],Val).

% optimize_program/2 and optimize/2 are used to reduce the number of
% calls to unif by grouping simple unifications. 
% ex: (X === [a:1,b:2], X === [c:3,d:4]) ---> (X === [a:1,b:2,c:3,d:4])
optimize_program([],[]).
optimize_program([(LHS:-RHS)|R],[(LHS:-ORHS)|RO]) :-
    !,
    conjList(RHS,LRHS),
    optimize(LRHS,OLRHS),
    listConj(OLRHS,ORHS),
    optimize_program(R,RO).
optimize_program([LHS|R],[LHS|RO]) :-
    optimize_program(R,RO).

optimize([],[]).
optimize([unif(X1,X2),unif(X3,X4)|R],RO) :-
    (V1=X1,V2=X3,S1=X2,S2=X4;V1=X2,V2=X3,S1=X1,S2=X4;
     V1=X1,V2=X4,S1=X2,S2=X3;V1=X2,V2=X4,S1=X1,S2=X3),
    var(V1),var(V2),nonvar(S1),nonvar(S2),V1 == V2,unif(S1,S2),
    !,optimize([unif(V1,S1)|R],RO).
optimize([X|R],[X|RO]) :- optimize(R,RO).

