%%%%%%%%%%%%%%%%%%%%%%%%%%%%% -*- Mode: Prolog -*- %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% tools.pl --- 
%% Author          : Massimo Fasciano
%% Last Modified On: Thu Oct 21 14:12:53 1993
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- ensure_loaded(library(strings)).

% build a new predicate with multiple alternate rules
% to replace an in-rule disjunction (;)
makeRules(Disj,Proc,Rules) :-
    listvars(Disj,Vars0),rev(Vars0,Vars),
    gensym('_',N),symbole(gram,N,ProcName),
    Proc =.. [ProcName|Vars],
    disjListRules(Disj,Proc,Rules).
disjListRules((A;B),P,[(CP:-CA)|R]) :-
    disjListRules(B,P,R),copy_term((P:-A),(CP :- CA)).
disjListRules(A,P,[(CP:-CA)]) :- copy_term((P:-A),(CP :- CA)).

% convert feature-level disjunction into rule-level disjunction
% ex: (X === (a;b;c)) ---> (X === a; X === b; X === c)
featureDisj((A===BorC),((A===B);R)) :-
    nonvar(BorC),BorC=(B;C),!,
    featureDisj((A===C),R).
featureDisj((A===B),(A===B)).

% get list of variables in a term. each variable appears only once.
list_of_vars(X,[X]) :- var(X),!.
list_of_vars(X,L) :- functor(X,_,A),!,list_of_vars_args(X,A,L).
list_of_vars(_,[]).
list_of_vars_args(_,0,[]) :- !.
list_of_vars_args(X,N,L) :-
    arg(N,X,A),list_of_vars(A,LA),N1 is N-1,
    list_of_vars_args(X,N1,LN1),append(LA,LN1,L).
listvars(X,L) :- list_of_vars(X,L0),listvars2(L0,L).
listvars2([],[]).
listvars2([H|T],[H|T2]) :- not vmember(H,T),!,listvars2(T,T2).
listvars2([H|T],T2) :- listvars2(T,T2).

% read a file as a list of terms
lire_fichier(L2) :- read(Terme),
    (Terme=end_of_file,L2=[],!;
     lire_fichier(L),L2=[Terme|L]).

% produce a conjunction of goals and eliminate "true" if possible.
conjTrue(true,true,true) :- !.
conjTrue(X,true,X) :- !.
conjTrue(true,X,X) :- !.
conjTrue(X,Y,(X,Y)) :- !.

% convert normal lists into open lists
% (with an unbound variable at the end).
% special case for the "pattern" feature
% ([x:1,y:2,pattern:[1,2]] ==> [x:1,y:2,pattern:[[1,2]|_]|_])
ouverte(A,A) :- var(A),!.
ouverte([],_) :- !.
ouverte([H|T],[(pattern:[Pat|_])|TT]) :-
    nonvar(H),H=(pattern:Pat),!,ouverte(T,TT).
ouverte([H|T],[(F:VV)|TT]) :-
    nonvar(H),H=(F:V),!,ouverte(V,VV),ouverte(T,TT).
ouverte([H|T],[H|TT]) :- !,ouverte(T,TT).
ouverte(A,A).

% for new symbol generation
symbole(B,N,S) :-
    name(B,RULE),name(N,NUM),append(RULE,NUM,RULENUM),name(S,RULENUM).

% print a list of rules (a program)
pr_prog([]).
pr_prog([H|T]) :-
    displayq(H),nl,
    pr_prog(T).

% display Prolog rules
writeQ(X) :- write(X).
displayq((H --> Tail)) :- !, writeQ(H), write(' -->'), nl, disptq(Tail).
displayq((H :- true)) :- !, writeQ(H), write('.'), nl.
displayq((H :- Tail)) :- !, writeQ(H), write(' :-'), nl, disptq(Tail).
displayq(A) :- writeQ(A), write('.'), nl.
disptq((A,B)) :- !, write('   '),writeQ(A), write(','), nl, disptq(B).
disptq(A) :- write('   '), writeQ(A), write('.'), nl.

% list concatenation
conc([],L,L).
conc([H|R1],L,[H|R2]) :-
    conc(R1,L,R2).

% member for variables
vmember(X,[Y|Ys]) :- X == Y,!.
vmember(X,[Y|Ys]) :- vmember(X,Ys).

% good old member
member(X,[X|_]).
member(Y,[X|Xs]) :- member(Y,Xs).

% reverse a list
rev(X,Y) :- rev(X,[],Y).
rev([],R,R).
rev([H|T],L,R) :-
    rev(T,[H|L],R).

% convert list to conjunction ([1,2,3] ---> (1,2,3))
listConj([H],H).
listConj([H|T],(H,TL)) :- listConj(T,TL).

% convert conjunction to list and flatten ((1,(2,3),4) ---> [1,2,3,4])
conjList((A,B),AB) :- !,conjList(A,A1),conjList(B,B1),conc(A1,B1,AB).
conjList(A,[A]).

